import traceback
import pytest
import os
from typing import Any, Callable, List, Optional, Type

from .exceptions import SnapshotDoesNotExist

from .io import SnapshotIO
from .serializer import SnapshotSerializer
from .location import TestLocation


class SnapshotAssertion:
    def __init__(
        self,
        *,
        update_snapshots: bool,
        io_class: Type[SnapshotIO],
        serializer_class: Type[SnapshotSerializer],
        test_location: TestLocation,
        session,
    ):
        self._update_snapshots = update_snapshots
        self._io_class = io_class
        self._serializer_class = serializer_class
        self._test_location = test_location
        self._executions = 0

        from .session import SnapshotSession

        self._session: SnapshotSession = session
        self._session.register_request(self)

    @property
    def io(self) -> SnapshotIO:
        if not getattr(self, "_io", None):
            self._io = self._io_class(
                test_location=self._test_location, file_hook=self._file_hook
            )
        return self._io

    @property
    def serializer(self) -> SnapshotSerializer:
        if not getattr(self, "_serializer", None):
            self._serializer = self._serializer_class()
        return self._serializer

    @property
    def num_executions(self) -> int:
        return int(self._executions)

    def with_class(
        self,
        io_class: Type[SnapshotIO] = None,
        serializer_class: Type[SnapshotSerializer] = None,
    ):
        return self.__class__(
            update_snapshots=self._update_snapshots,
            test_location=self._test_location,
            io_class=io_class or self._io_class,
            serializer_class=serializer_class or self._serializer_class,
            session=self._session,
        )

    def assert_match(self, data):
        assert self == data

    def get_assert_diff(self, data) -> List[str]:
        deserialized = self._recall_data(index=self.num_executions - 1)
        if deserialized is None:
            return ["Snapshot does not exist!"]

        if data != deserialized:
            return [f"- {data}", f"+ {deserialized}"]

        return []

    def _file_hook(self, filepath, snapshot_name):
        self._session.add_visited_snapshots({filepath: {snapshot_name}})

    def __repr__(self) -> str:
        return f"<SnapshotAssertion ({self.num_executions})>"

    def __call__(self, data) -> bool:
        return self._assert(data)

    def __eq__(self, other) -> bool:
        return self._assert(other)

    def _assert(self, data) -> bool:
        self._session.register_assertion(self)
        try:
            if self._update_snapshots:
                serialized_data = self.serializer.encode(data)
                self.io.create_or_update_snapshot(
                    serialized_data, index=self.num_executions
                )
                return True

            deserialized = self._recall_data(index=self.num_executions)
            if deserialized is None or data != deserialized:
                return False
            return True
        finally:
            self._executions += 1

    def _recall_data(self, index: int) -> Optional[Any]:
        try:
            saved_data = self.io.read_snapshot(index=index)
            return self.serializer.decode(saved_data)
        except SnapshotDoesNotExist:
            return None



from .io import AbstractImageSnapshotIO
from .serializer import ImageSnapshotSerializer


class PNGImageSnapshotIO(AbstractImageSnapshotIO):
    @property
    def extension(self) -> str:
        return "png"


class SVGImageSnapshotIO(AbstractImageSnapshotIO):
    @property
    def extension(self) -> str:
        return "svg"



from typing import Any, Set, Optional
from abc import ABC, abstractmethod
import re

import os

from syrupy.io import SnapshotIO
from syrupy.exceptions import SnapshotDoesNotExist
from syrupy.types import SnapshotFiles


class AbstractImageSnapshotIO(ABC, SnapshotIO):
    @property
    @abstractmethod
    def extension(self) -> str:
        pass

    def discover_snapshots(self, filepath: str) -> Set[str]:
        return {os.path.splitext(os.path.basename(filepath))[0]}

    def get_file_basename(self, index: int) -> str:
        maybe_extension = f".{self.extension}" if self.extension else ""
        sanitized_name = self._clean_filename(self.get_snapshot_name(index=index))
        return f"{sanitized_name}{maybe_extension}"

    def _get_snapshot_dirname(self):
        return os.path.splitext(os.path.basename(str(self.test_location.filename)))[0]

    def _read_snapshot_from_file(self, snapshot_file: str, _):
        return self._read_file(snapshot_file)

    def _read_file(self, filepath: str) -> Any:
        try:
            with open(filepath, "rb") as f:
                return f.read()
        except FileNotFoundError:
            return None

    def _write_snapshot_or_remove_file(self, snapshot_file: str, _: str, data: Any):
        if data:
            self._write_file(snapshot_file, data)
        else:
            os.remove(snapshot_file)

    def _write_file(self, filepath: str, data: Any):
        with open(filepath, "wb") as f:
            f.write(data)

    def _clean_filename(self, filename: str) -> str:
        filename = str(filename).strip().replace(" ", "_")
        max_filename_length = 255 - len(self.extension or "")
        return re.sub(r"(?u)[^-\w.]", "", filename)[:max_filename_length]



import os

import semver
from invoke import task


@task
def clean(ctx):
    ctx.run(f"rm -rf *.egg-info dist build __pycache__ .pytest_cache artifacts/*")


@task
def requirements(ctx):
    ctx.run(f"python -m piptools compile dev-requirements.in", pty=True)


@task
def lint(ctx, fix=False):
    ctx.run(
        f"python -m black --target-version py36 {'--check' if not fix else ''} ./*.py ./src ./tests",
        pty=True,
    )

    if fix:
        print("\nSkipping type check as there is no fixer")
    else:
        print("\nRunning type check")
        ctx.run("python -m mypy --ignore-missing-imports src", pty=True)


@task
def test(ctx, update_snapshots=False, verbose=False):
    ctx.run(
        "python -m pytest ."
        f"{' -s' if verbose else ''}"
        f"{' --update-snapshots' if update_snapshots else ''}",
        env={"PYTHONPATH": "./src"},
        pty=True,
    )


@task(pre=[clean])
def build(ctx):
    ctx.run("python setup.py sdist bdist_wheel")


@task
def publish(ctx, dry_run=True):
    ctx.run(
        f"twine upload {'--repository-url https://test.pypi.org/legacy/' if dry_run else ''} dist/*"
    )


@task(pre=[build])
def release(ctx, dry_run=True):
    if not dry_run and not os.environ.get("CI"):
        print("This is a CI only command")
        exit(1)

    # get version created in build
    with open("version.txt", "r") as f:
        version = str(f.read())

    try:
        should_publish_to_pypi = not dry_run and semver.parse_version_info(version)
    except ValueError:
        should_publish_to_pypi = False

    # publish to test to verify builds
    publish(ctx, dry_run=True)

    # publish to pypi if test succeeds
    if should_publish_to_pypi:
        publish(ctx, dry_run=False)



def red(text):
    return f"\033[31m{text}\033[0m"


def yellow(text):
    return f"\033[33m{text}\033[0m"


def bold(text):
    return f"\033[1m{text}\033[0m"


def error_style(text):
    return bold(red(text))
